#!/usr/bin/env python3 
'''
Created on Mar 23, 2017

@author: arnon
'''
from openpyxl import load_workbook
import collections
import os

def get_range(table_range=None, sheet_name=None):
    if sheet_name is not None and isinstance(table_range, collections.Mapping):
            table_range=table_range.get(sheet_name)
            
    if table_range is None or len(table_range) == 0:
        table_range=((1,1), (0,0))
    start_coor, end_coor = table_range
    if len(start_coor) == 0:
        start_coor=(1,1)
    start_row, start_col=start_coor
    if len(end_coor) > 0:
        end_coor=(0,0)
    end_row, end_col=end_coor
    return start_row, start_col, end_row, end_col

def get_sep_line(sym='-', sizes=[]):
    content=[sym*s for s in sizes]
    return '+' + '+'.join(content) + '+' 

def cell2str(cell):
    value=cell.value
    if value is None: value=''
    return str(value)

def mkline(row, col_size):
    line_items=[ "%s%s" %(cell2str(cell), " "*(size-len(cell2str(cell))))for cell, size in zip(row, col_size)]
    return '|' + '|'.join(line_items) + '|\n'

def get_dest_base(file, dest=None, one_file=True):
    if dest is None:
        result=file.replace('.xlsx', '')
    else:
        result=dest.replace('.rst', '')
    return result

def xlsx2rst(file, dest=None, sheets=[], one_file=True, table_range=None, header_row=1, ):
    ''' converts xlsx to rst format
    
    Args:
        file: path to xlsx file
        sheet: if empty, all sheets will be converted.  otherwise, only specified sheet-names will be produced.
            note: sheets names are case sensitive.
        table_range: start end coordinates in the form of ((row1, col1), (row2, col2))
            if (row1, col1) is empty, (1,1) is assumed
            if (row2, col2) is empty, last field filled
        header_row: number of row off range to take as header
        one_file: if True, will produce one file with all sheet tables consolidate separated by sheet-name title.
            if False, a file per sheet will be produced.
        dest: path to destination, defaults to xlsx.rst.  if dest is not None and one_file is False, dest would be extended with sheet name.
    '''
    
    if not file.endswith('.xlsx'): raise Exception("file must ends with .xlsx: %s" % file)
    if dest is not None and not dest.endswith('.rst'): raise Exception("dest must ends with .rst: %s" % dest)
    wb = load_workbook(file)
    wbsheets=wb.get_sheet_names()
    
    if sheets is not None and len(sheets) > 0:
        src_sheets=list(set(wbsheets) & set(sheets))
        if len(src_sheets) == 0:
            raise Exception("sheets not found with workbook: %s" % sheets)
    else:
        src_sheets=wbsheets
        
    if len(src_sheets) == 0:
        raise Exception("workbook does not have sheets")
    
    dest_base=get_dest_base(file, dest, one_file)
    
    if one_file:
        dest=dest_base+'.rst'
        if os.path.isfile(dest):
            os.remove(dest)
        
    for sheet_name in src_sheets:
        # sheet=wb.get_sheet_by_name(first_sheet)
        ws=wb[sheet_name]
        start_row, start_col, end_row, end_col= get_range(table_range, sheet_name)
        row_count=0
        header=list()
        body=list()
        col_size=list()
        
        if not one_file:
            dest=dest_base+'.%s.rst' % sheet_name
            if os.path.isfile(dest):
                os.remove(dest)
        
        with open(dest, 'a') as destfile:
            for row in ws.iter_rows():
                row_count += 1
                if row_count < start_row: continue
                if end_row > 0 and row_count > end_row: break
                cell_size=[len(str(cell.value)) for cell in row]
                if len(col_size) > 0:
                    col_size=[max(cols, cells) for cols, cells in zip(col_size, cell_size)]
                else:
                    col_size=cell_size  
                     
                eff_row=row_count-start_row+1
                
                if eff_row <= header_row:
                    header.append(row)
                else:
                    body.append(row)
                    
            body_sep=get_sep_line(sym='-', sizes=col_size)
            header_sep=get_sep_line(sym='=', sizes=col_size)
            
            destfile.write("%s\n%s\n\n" % (sheet_name, '='*len(sheet_name)) + '\n')
            destfile.write(body_sep+'\n')
            for row in header:
                destfile.write(mkline(row, col_size=col_size))
            destfile.write(header_sep+'\n')
            for row in body:
                destfile.write(mkline(row, col_size=col_size))
                destfile.write(body_sep+'\n')
                
if '__main__' == __name__:
    import argparse
    parser = argparse.ArgumentParser(description='Converts xlsx workbook into restructured text format')
    parser.add_argument("file", metavar='XLSX', 
                        help="xlsx files to convert")
    parser.add_argument("-o", "--output", metavar='RST', required=False, 
                        help="destination rst file")
    parser.add_argument("-s", "--sheet", metavar='SHEET', nargs='*', 
                        help="list of sheets; default to all available sheets")
    parser.add_argument("--start-row", metavar='NUMBER', type=int, nargs='?', default=1, dest='start_row',
                        help="table start row, defaults to 1")
    parser.add_argument("--end-row", metavar='NUMBER', type=int, nargs='?', default=1, dest='start_col',
                        help="table start col, defaults to 1")
    parser.add_argument("--start-col", metavar='NUMBER', type=int, nargs='?', default=0, dest='end_row',
                        help="table start row, defaults to 0")
    parser.add_argument("--end-col", metavar='NUMBER', type=int, nargs='?', default=0, dest='end_col',
                        help="table start col, defaults to 0")
    parser.add_argument("-r", "--header", metavar='NUMBER', type=int, nargs='?', default=1,
                        help="header row count")
    parser.add_argument("--one-file", action='store_true', default=False,
                        help="when set, single file is created")
    
    args = parser.parse_args()
    
    table_range=((args.start_row, args.start_col), (args.end_row, args.end_col))
    xlsx2rst(file=args.file, dest=args.output, sheets=args.sheet, table_range=table_range, header_row=args.header, one_file=args.one_file)


#!/usr/bin/env python3 
# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

'''
# based on Recipe 466302: Sorting big files the Python 2.4 way
# by Nicolas Lehuen
# fixed for python 3.x

@author: arnon
'''

import os
from tempfile import gettempdir
from itertools import islice, cycle
from collections import namedtuple
import heapq

Keyed = namedtuple("Keyed", ["key", "obj"])

def merge(key=None, *iterables):
    # based on code posted by Scott David Daniels in c.l.p.
    # http://groups.google.com/group/comp.lang.python/msg/484f01f1ea3c832d

    if key is None:
        keyed_iterables = iterables
    else:
        keyed_iterables = [(Keyed(key(obj), obj) for obj in iterable)
                            for iterable in iterables]
    for element in heapq.merge(*keyed_iterables):
        yield element.obj


def msort(input_, output, key=None, buffer_size=32000, tempdirs=None):
    if tempdirs is None:
        tempdirs = []
    if not tempdirs:
        tempdirs.append(gettempdir())

    chunks = []
    try:
        with open(input_,'rb',64*1024) as input_file:
            input_iterator = iter(input_file)
            for tempdir in cycle(tempdirs):
                current_chunk = list(islice(input_iterator,buffer_size))
                if not current_chunk:
                    break
                current_chunk.sort(key=key)
                output_chunk = open(os.path.join(tempdir,'%06i'%len(chunks)),'w+b',64*1024)
                chunks.append(output_chunk)
                output_chunk.writelines(current_chunk)
                output_chunk.flush()
                output_chunk.seek(0)
        with open(output,'wb',64*1024) as output_file:
            output_file.writelines(merge(key, *chunks))
    finally:
        for chunk in chunks:
            try:
                chunk.close()
                os.remove(chunk.name)
            except Exception:
                pass


if __name__ == '__main__':
    import argparse
    
    parser = argparse.ArgumentParser(description="""
Sort big data file.
""")
    parser.add_argument('-b', '--buffer', required=True, type=int, metavar='MEMSIZE', default=32000, dest='buffer_size',
                        help="""Size of the line buffer. The file to sort is divided into chunks of that many lines. Default : 32,000 lines.""")
    parser.add_argument('-k', '--key', type=str, required=False, metavar='KEY', dest='key',
                        help="""Python expression used to compute the key for each line, "lambda line:" is prepended. 
                                Example : -k "line[5:10]". By default, the whole line is the key.""")    
    parser.add_argument('-t', '--tempdir', action='append', metavar='DIR', default=[], dest='tempdirs',
                        help="""Temporary directory to use. You might get performance
                                improvements if the temporary directory is not on the same physical
                                disk than the input and output directories. You can even try
                                providing multiples directories on differents physical disks.
                                Use multiple -t options to do that.""")
    parser.add_argument('-p', '--psyco', action='store_true', required=False, default=False, dest='psyco',
                        help="""Use Psyco.""")
    parser.add_argument('input', type=str, metavar='INPUT',
                        help="""File to sort.""")
    parser.add_argument('output', type=str, metavar='OUTPUT',
                        help="""Sorted file to create.""")

    args = parser.parse_args()    
    
    if args.key:
        args.key = eval('lambda line : (%s)'% args.key)

    if args.psyco:
        try:
            import psyco
        except:
            raise Exception("psyco package could not be loaded.  To use -p option, psyco must be installed.")
        psyco.full()

    msort(args.input, args.output, args.key, args.buffer_size, args.tempdirs)
    




#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

"""Send the contents of a directory as a MIME message."""

import os
import sys
import smtplib
# For guessing MIME type based on file name extension
import mimetypes
from contextlib import contextmanager

from argparse import ArgumentParser

from email import encoders
from email.message import Message
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

COMMASPACE = ', '

def attach(outer, path):
    # Guess the content type based on the file's extension.  Encoding
    # will be ignored, although we should check for simple things like
    # gzip'd or compressed files.
    ctype, encoding = mimetypes.guess_type(path)
    if ctype is None or encoding is not None:
        # No guess could be made, or the file is encoded (compressed), so
        # use a generic bag-of-bits type.
        ctype = 'application/octet-stream'
    maintype, subtype = ctype.split('/', 1)
    if maintype == 'text':
        with open(path) as fp:
            # Note: we should handle calculating the charset
            msg = MIMEText(fp.read(), _subtype=subtype)
    elif maintype == 'image':
        with open(path, 'rb') as fp:
            msg = MIMEImage(fp.read(), _subtype=subtype)
    elif maintype == 'audio':
        with open(path, 'rb') as fp:
            msg = MIMEAudio(fp.read(), _subtype=subtype)
    else:
        with open(path, 'rb') as fp:
            msg = MIMEBase(maintype, subtype)
            msg.set_payload(fp.read())
        # Encode the payload using Base64
        encoders.encode_base64(msg)
    # Set the filename parameter
    msg.add_header('Content-Disposition', 'attachment', filename=path)
    outer.attach(msg)

@contextmanager
def activate(connection):
    protocol=connection.get('protocol', '')
    host=connection.get('host', os.environ.get('SMTP_HOST', 'localhost'))
    port=connection.get('port', int(os.environ.get('SMTP_PORT', '0')))
    username=connection.get('username', os.environ.get('SMTP_USERNAME', os.environ.get('USER')))
    password=connection.get('password', os.environ.get('SMTP_PASSWORD', ''))
    
    if not protocol:
        from smtplib import SMTP as SMTP
    elif protocol=='ssl':
        from smtplib import SMTP_SSL as SMTP
    else:
        raise Exception("Unknown protocol: 's'" % (protocol,))
    
    smtp_params={'host': host}
    if port:
        smtp_params['port']=port
    elif not protocol:
        smtp_params['port']=110
    elif protocol=='ssl':
        smtp_params['port']=465
    
    try:
        smtp_srv = SMTP(**smtp_params)
    except Exception as err:
        raise Exception('Failed to initiate SMTP: {}; SMTP PARAMS: {}'.format(repr(err), repr(smtp_params)))

    if protocol == 'ssl':
        try:
            smtp_srv.login(username, password)
        except Exception as err:
            smtp_srv.close()
            raise Exception("Failed to login to SMTP server: {}: {}".format(username, repr(err)))

    yield smtp_srv
    smtp_srv.close()


def send_mail(mailto, subject, body='', mailfrom=None, mailcc=[], mailbcc=[], attachments=[], output='', connection={}):
    ''' sends mail with attachements
        
        Args:
        mailto: (list) of email addresses
        subject: (str) text of email subject
        body: (str) text of email body; default empty text
        mailfrom: (list) email address to use as from address; default to ${USER}@${HOST}
        mailcc: (list) addresses to send carbon copy to
        mailbcc: (list) addresses to send blind carbon copy to
        attachments: (list) path to files to attach to mail
        output: (str) path to wrtie email format to instead of sending
        connection: (dict) connection settings to use when sending email it should have the following keys:
            'protocol': 'ssl', regular SMTP will be used if not provided.
            'username': name to use to login to SMPT server. $SMTP_USERNAME or $USER is used if not provided.
            'password': password to use to connect to to SMTP server. $SMTP_PASSWORD if not provided.
            'host': (str) url of SMTP server. SMTP_HOST or local host is used if not provided.
            'port': (int) port number to use.  defaults to SMTP_PORT or 110 for SMTP and 465 for SSL
        
        '''
    
    # Create the enclosing (outer) message
    outer = MIMEMultipart()
    if not subject:
        subject = 'Content of directory'
    
    outer['Subject'] = subject
    if mailcc:
        outer['Cc'] = COMMASPACE.join(mailcc)
    if mailbcc:
        outer['BCc'] = COMMASPACE.join(mailbcc)

    outer['To'] = COMMASPACE.join(mailto)

    if mailfrom is None:
        mailfrom="{}@{}".format(os.environ['USER'], os.environ['HOST'])
    outer['From'] = mailfrom

    outer.preamble = 'You will not see this in a MIME-aware mail reader.\n'
    if body:
        outer.attach(MIMEText(body))

    for attachment in attachments:
        if not os.path.isfile(attachment):
            for filename in os.listdir(attachment):
                path = os.path.join(attachment, filename)
                if not os.path.isfile(path):
                    continue
                attach(outer, path)
        else:
            attach(outer, attachment)
    # Now send or store the message
    composed = outer.as_string()
    if output:
        with open(output, 'w') as fp:
            fp.write(composed)
    else:
        with activate(connection) as s:
            s.send_message(outer)

def get_connection(ssl=False, connection=None):
    '''
        'protocol': 'ssl', regular SMTP will be used if not provided.
        'username': name to use to login to SMPT server. $SMTP_USERNAME or $USER is used if not provided.
        'password': password to use to connect to to SMTP server. $SMTP_PASSWORD if not provided.
        'host': (str) url of SMTP server. SMTP_HOST or local host is used if not provided.
        'port': (int) port number to use.  defaults to SMTP_PORT or 110 for SMTP and 465 for SSL

    '''
    result=dict()
    
    if ssl:
        result['protocol']='ssl'
    
    if connection:
        'user:pass@host:port'
        mparts=connection.rpartition('@')
        if mparts[0]:
            parts=mparts[0].partition(':')
            if parts[0]:
                result['username']=parts[0]
            if parts[2]:
                result['password']=parts[2]
        if mparts[2]:
            parts=mparts[2].partition(':')
            if parts[0]:
                result['host']=parts[0]
            if parts[2]:
                result['port']=parts[2]
    return result

if __name__ == '__main__':
    parser = ArgumentParser(description="""\
        Send the contents of a directory as a MIME message.
        Unless the -o option is given, the email is sent by forwarding to your local
        SMTP server, which then does the normal delivery process.  Your local machine
        must be running an SMTP server.
        """)
    parser.add_argument('-a', '--attach', required=False,
                        action='append', metavar='ATTACHMENT', default=[], dest='attachments',
                        help="""Mail the contents of the specified directory or file, Only the regular
                            files in the directory are sent, and we don't recurse to
                            subdirectories.""")
    parser.add_argument('-o', '--output', metavar='FILE',
                        help="""Print the composed message to FILE instead of
                            sending the message to the SMTP server.""")
    parser.add_argument('-s', '--subject', metavar='SUBJECT', required=True,
                        help="""Subject for email message (required).""")
    parser.add_argument('-b', '--body', metavar='BODY', required=False,
                        help="""Boby text for the message (optional).""")
    parser.add_argument('-f', '--mailfrom', required=False,
                        help='The value of the From: header (optional); if not provided $USER@$HOSTNAME will be use as sender')
    parser.add_argument('-c', '--malicc', required=False,
                        help='The value of the CC: header (optional)', action='append', metavar='CC',
                                            default=[], dest='mailcc')
    parser.add_argument('-t', '--mailto', required=True,
                        action='append', metavar='RECIPIENT',
                        default=[], dest='mailto',
                        help='A To: header value (at least one required)')    
    parser.add_argument('--ssl', action='store_true', dest='ssl',
                        help="""Sets mail to use SSL protocol""")

    parser.add_argument('-C', '--connection', metavar='CONNECTION', dest='connection',
                        help="""provide connection URI: [user[:pass]][@[host[:port]]]""")

    args = parser.parse_args()
    
    connection=get_connection(args.ssl, args.connection)
    send_mail(subject=args.subject, body=args.body, mailfrom=args.mailfrom, mailto=args.mailto, mailcc=args.mailcc, attachments=args.attachments, output=args.output, connection=connection)
